/*
 * File:      cgpltcl.c
 * Purpose:   Implements a somewhat primitive GNU Prolog/Tcl interface
 * Project:   AAAAA version 11
 * Author:    Anders Andersson
 * Created:   July 22, 1999
 * Modified:  July 26, 1999
 * Notes:     Uses Mktclapp and the GNU Prolog foreign function interface.
 *
 * Copyright (c) Anders Andersson 1999
 *
 * May be redistributed under the GNU Public Licence.
 */

#include <stdlib.h>
#include <string.h>
#include "a5.h"
#include "gprolog.h"

#define ENTER /* fprintf(stderr,"<\n") */
#define EXIT /* fprintf(stderr,">\n") */

/* The queue we use to keep track of the pending events */

struct prolog_event_queue {
  char *event;
  struct prolog_event_queue *next;
};

static struct prolog_event_queue *queue_start = 0;
static struct prolog_event_queue *queue_end = 0;

/* This is the last string to be dequeued but not yet freed. */
static char *current_string = 0;

/* This is the last string to be protected but not yet freed */
static char *last_protected = 0;

/* Add a string to the end of the queue, returns 1 if OK, 0 if alloc fails */
static int enqueue(char *str)
{
  struct prolog_event_queue *newcons;
  char *newstring;
  int len;

  newcons = (struct prolog_event_queue *)
    malloc(sizeof(struct prolog_event_queue));
  if(newcons == 0) {
    return 0;
  }
  len = strlen(str);
  newstring = (char *)malloc(len+2);
  if(newstring == 0) {
    free(newcons);
    return 0;
  }
  strcpy(newstring, str);
  newstring[len] = '.';
  newstring[len+1] = 0;
  newcons->event = newstring;
  newcons->next = 0;
  if(queue_end == 0) queue_start = newcons;
  else queue_end->next = newcons;
  queue_end = newcons;
  return 1;
}

/* Free the current string, and set it to NULL */
static void free_current_string(void)
{
  if(current_string != 0) free(current_string);
  current_string = 0;
}

/* Free the last protected string, and set it to NULL */
static void free_last_protected(void)
{
  if(last_protected != 0) Tcl_Free(last_protected);
  last_protected = 0;
}

/* Remove the head of the queue, free the old current string, move the
   string from the head to the current string, and return it. 
   Returns NULL if the queue is empty, in which case the current string
   remains unchanged */
static char *dequeue(void)
{
  struct prolog_event_queue *head;

  head = queue_start;
  if(head == 0) return 0;  /* OK, so there is nothing on the queue */
  else {
    queue_start = queue_start->next;
    if(queue_start == 0) queue_end = 0; /* removing only element */
    free_current_string();
    current_string = head->event;
    free(head);
    return current_string;
  }
}

/* Check if queue is non-empty */
int prolog_event_pending(void)
{
  int res;
  res = (queue_start != 0);
  return res; 
}

/* Return the current string -- prologize if there isn't one */
char *peek_current_string(void)
{
  char *res;
  ENTER;
  if(current_string == 0) res = "[]";
  else res = current_string;
  EXIT;
  return res;
}

/* Implements the Tcl command "prolog_event EVENT" */
int ET_COMMAND_prolog_event(ET_TCLARGS)
{
  int res;
  ENTER;
  if(argc != 2) {
    Et_ResultF(interp,"wrong # args: should be \"%s EVENT\"", argv[0]);
    res = TCL_ERROR;
  }
  else if(enqueue(argv[1]))
    res = TCL_OK;
  else {
    Et_ResultF(interp,"out of memory");
    res = TCL_ERROR;
  }
  EXIT;
  return res;
}

/* Implements the Prolog predicate get_raw_event/1 which returns a Prolog
   string (list of character atoms). Fails if no event is pending. */
Bool get_raw_event(char **str)
{
  Bool res;
  ENTER;
  *str = dequeue();
  if(*str != 0)
    res = TRUE;
  else
    res = FALSE;
  EXIT;
  return res;
}

/* Implements the Prolog predicate raw_tcl_eval/3 which is given a Prolog
   string (list of character atoms) to execute and returns a status indicator
   (one of the atoms ok, error, return, break, continue, or unknown) and the 
   result (as another Prolog string). */
Bool raw_tcl_eval(char *string, char **status, char **result)
{
  int res;
  ENTER;
  res = Tcl_GlobalEval(Et_Interp, string);
  switch(res)
    {
    case TCL_OK:       *status = "ok";       break;
    case TCL_ERROR:    *status = "error";    break;
    case TCL_RETURN:   *status = "return";   break;
    case TCL_BREAK:    *status = "break";    break;
    case TCL_CONTINUE: *status = "continue"; break;
    default:           *status = "unknown";
    }
  *result = Et_Interp->result;
  EXIT;
  return TRUE;
}

/* Implements the Prolog predicate protect_tcl_chars/2 which is given a list
   of character codes corresponding to a raw string, and returns a Prolog
   string (list of character atoms) corresponding to Tcl quoted version. 
   If we run out of memory, we silently fail. */
Bool protect_tcl_chars(char *string, char **result)
{
  Bool res;
  ENTER;
  free_last_protected();
  last_protected = mprintf("\"%q\"", string);
  if(last_protected != 0) {
    *result = last_protected;
    res = TRUE;
  }
  else res = FALSE;
  EXIT;
  return res;
}



