/*
 * File:    gc_c.c
 * Purpose: A simple and incomplete garbage collector for GNU Prolog
 * Project: AAAAA version 11
 * Author:  Anders Andersson
 *
 * Copyright (c) Anders Andersson 1999.
 *
 * May be redistributed under the GNU General Public Licence.
 *
 */

/* ALGORITHM

The algorithm used is kind of a bastard between a mark-sweep and a coying
collector. Its performance characteristics mostly resemble a copying collector,
except that there is a small scan overhead (about a millisecond per 5 MB).

First we mark the heap in a traditional way, but there is no need to bother 
about pointer reversal or other optimizations, since we are wasting memory 
for the later phases anyway. During this phase, also set a bit indicating
if a word is a pointer into the heap or not.

Second, we compact the data into "tospace" (which isn't tospace at all really,
it is just a buffer... It was tospace in an earlier version, which saved the
moving at the end, but it had some disadvantages). This is done by scanning 
the marked words, copying each one into tospace and overwriting it with a
index to its new location. We also compact the bits indicating if it is 
a pointer or not. 

Third, we adjust the pointers. If a value in tospace is flagged as a pointer,
replace it with heap_start + index, where it points to a word overwritten
with index. Also do this with all the root pointers. Pointers into the heap
that are not root pointers, such as the saved H's in choice-points are also 
updated, by a linear search down to the last marked word.

Fourth, we copy tospace back into the main heap. 

*/

#include <stdio.h>
#include <string.h>

#include "gprolog.h"
#include "gc_c.h"

/* The number of WamWords that might be allocated between calls to GC. */

/* If user gives 0, then we use this: */
#define DEFAULT_MARGIN 1024

/* And if user gives -1, then we use this: (20 means 1/20 = 5%) */
#define DEFAULT_MARGIN_FRACTION 20

/* The default value for max_fill_percent */
#define DEFAULT_FILL 75

/* The maximum number of nested catches ... */
#define MAX_NUM_CATCHES 1000

static int margin;

/* ... the stack of catches ... */
static WamWord *old_memory_catches[MAX_NUM_CATCHES];

/* ... and the stack pointer ... */
static int num_catches;

/* Unsigned version of the WamWord, used for bitmasks. */
typedef unsigned long UWW;

/* Used to count the number of marked words */
static unsigned long count;

/* Used to decide when we run out of memory */
static unsigned long max_count;

/* Used to keep track of the previous heap marker during forwarding */
static int not_above;

/* The scope of the heap */
static WamWord *heap_start, *heap_trigger_gc, *heap_trigger_out, *heap_end;

/* Various places in tospace */
static WamWord *tospace,*lis_sp, *stc_sp, *movingbits, *markbits, *tospace_end;

/* Used to cut when we run out of memory. I can't seem to get throw to work */
static WamWord *memory_catch_B;

/* Defines log_2 WORD_SIZE */
#if WORD_SIZE==32
#define LOG_WS 5
#else
#define LOG_WS 6
#endif

/* Convert a pointer to an index into the heap. As Global_Offset but faster. */
#define PointerToIndex(adr) ((WamWord *)(adr)-heap_start)

/* From an index into the heap, compute the index into the mark array */
#define IndexToMarkWord(idx) ((idx) >> LOG_WS)

/* From an index into the heap, compute the bitmask for the mark array bit # */
#define IndexToMarkIndex(idx) (1 << ((idx) & (WORD_SIZE-1)))

/* Set a mark bit */
#define Mark(mw,mi) do{markbits[mw] |= (mi);count++;}while(0)

/* Check a mark bit */
#define IsMarked(mw,mi) (markbits[mw] & (mi))

/* Set the bit indicating that something is a heap pointer */
#define MarkMoving(mw,mi) movingbits[mw] |= (mi)

/* Clear the bit indicating that something is a heap pointer */
#define MarkNonMoving(mw,mi) movingbits[mw] &= ~(mi)

/* Check if the heap pointer bit is set */
#define IsMoving(mw,mi) (movingbits[mw] & (mi))

/* Push a pointer to the stack holding lists */
#define Push_LST(p) (*lis_sp++ = (WamWord) p)

/* Pop a pointer from the stack holding lists */
#define Pop_LST() ((WamWord *)(*--lis_sp))

/* Check if there is anything on the stack holding lists */
#define LST_Stack_Inhabited() (lis_sp != tospace)

/* Push a pointer to the stack holding structures */
#define Push_STC(p) (*--stc_sp = (WamWord) p)

/* Pop a pointer from the stack holding structures */
#define Pop_STC() ((WamWord *)(*stc_sp++))

/* Check if there is anything on the stack holding structures */
#define STC_Stack_Inhabited() (stc_sp != movingbits)

/* Given a pointer to a float on the heap, mark the words */
static void Mark_Float(WamWord *p)
{
  int i, mw;
  UWW mi;
  i = PointerToIndex(p);
  mi = IndexToMarkIndex(i);
  mw = IndexToMarkWord(i);
  if(IsMarked(mw,mi)) return;
  Mark(mw,mi);
  MarkNonMoving(mw,mi);
  #if WORD_SIZE==32
  i++;
  mi = IndexToMarkIndex(i);
  mw = IndexToMarkWord(i);
  Mark(mw,mi);
  MarkNonMoving(mw,mi);
  #endif
}

/* Given an index of a interior word in a structure or cons, and the contents
   of that word, and assuming the word itself has just been marked for the
   first time. Mark any chain as needed and push LST or STR as needed, and also
   mark it as movable or not.
*/
static void Mark_Interior(WamWord w, int mw, UWW mi)
{
  int i;
  WamWord *p;
 start:
  switch(Tag_Of(w)) 
    {
    case REF:
      p = UnTag_Stack(w);
      if(p < heap_start || p >= heap_end) {
	MarkNonMoving(mw,mi);
	break;
      }
      MarkMoving(mw,mi);
      w = *p;
      i = PointerToIndex(p);
      mi = IndexToMarkIndex(i);
      mw = IndexToMarkWord(i);
      if(IsMarked(mw,mi)) break;
      Mark(mw, mi);
      goto start;
    case FLT:
      p = UnTag_Stack(w);
      if(p < heap_start || p >= heap_end) {
	MarkNonMoving(mw,mi);
	break;
      }
      MarkMoving(mw,mi);
      Mark_Float(p);
      break;
    case LST:
      p = UnTag_Stack(w);
      if(p < heap_start || p >= heap_end) { 
	MarkNonMoving(mw,mi);
	break;
      }
      MarkMoving(mw,mi);
      Push_LST(p);
      break;
    case STC:
      p = UnTag_Stack(w);
      if(p < heap_start || p >= heap_end) { 
	MarkNonMoving(mw,mi);
	break;
      }
      MarkMoving(mw,mi);
      Push_STC(p);
      break;      
    default: 
      MarkNonMoving(mw,mi);
      break;
    }
}

/* Given a pointer to a cons cell, mark the CAR and CDR w/ Mark_Interior */
static void Mark_List(WamWord *p)
{
  int i, mw;
  UWW mi;
  WamWord w;
  i = PointerToIndex(p);
  mi = IndexToMarkIndex(i);
  mw = IndexToMarkWord(i);
  if(!IsMarked(mw, mi)) {  /* Note: CAR may be marked even if CDR is not */
    Mark(mw,mi);
    w = *p;
    Mark_Interior(w, mw, mi);
  }
  i++;
  mi = IndexToMarkIndex(i);
  mw = IndexToMarkWord(i);
  if(!IsMarked(mw, mi)) {
    Mark(mw,mi);
    w = p[1];
    Mark_Interior(w, mw, mi);
  }
}

/* Given a pointer to the start of a structure, mark it w/ Mark_Interior */
static void Mark_Structure(WamWord *p)
{
  int i, mw, n;
  UWW mi;
  WamWord w;
  i = PointerToIndex(p);
  mi = IndexToMarkIndex(i);
  mw = IndexToMarkWord(i);
  if(IsMarked(mw, mi)) return;  /* If the functor-field is marked, then so
				   is the entire structure */
  Mark(mw,mi);
  MarkNonMoving(mw,mi);
  w = *p;
  n = Arity_Of(w);
  while(n-- > 0) {
    i++;
    p++;
    mi = IndexToMarkIndex(i);
    mw = IndexToMarkWord(i);
    if(!IsMarked(mw,mi)) {
      Mark(mw,mi);
      w = *p;
      Mark_Interior(w, mw, mi);
    } 
  }
}

/* Given a WamWord outside the heap which may be a root pointer, mark 
 * everything that is accessible from it */
static void Mark_Root(WamWord w)
{
  WamWord *p;
  int i, mw;
  UWW mi;

  switch(Tag_Of(w))
    {
    case REF:
      p = UnTag_Stack(w);     
      if(p < heap_start || p >= heap_end) break;
      i = PointerToIndex(p);
      mi = IndexToMarkIndex(i);
      mw = IndexToMarkWord(i);
      if(IsMarked(mw,mi)) break;
      Mark(mw,mi);
      w = *p;
      Mark_Interior(w, mw, mi);
      break;
    case FLT:
      p = UnTag_Stack(w);
      if(p < heap_start || p >= heap_end) break;
      Mark_Float(p);
      break;
    case LST:
      p = UnTag_Stack(w);
      if(p < heap_start || p >= heap_end) break;
      Mark_List(p);
      break;
    case STC:
      p = UnTag_Stack(w);
      if(p < heap_start || p >= heap_end) break;
      Mark_Structure(p);
      break;
    }
  for(;;) {
    if(LST_Stack_Inhabited()) Mark_List(Pop_LST());
    else if(STC_Stack_Inhabited()) Mark_Structure(Pop_STC());
    else return;
  }
}

/* Given a pointer to a WamWord outside the heap which may be a root pointer, 
 * forward it if it indeed points to the heap. */
static WamWord Forward_Root(WamWord w)
{
  int tag;
  WamWord *p, *p2;
  tag = Tag_Of(w);
  switch(tag)
    {
    case REF:
    case FLT:
    case LST:
    case STC:
      p = UnTag_Stack(w);
      if(p < heap_start || p >= heap_end) return w;
      p2 = heap_start + *p;
      return Tag_Value(tag, p2);
    default:
      return w;
    }
}

/* Go through the words copied to tospace and forward pointers to fromspace */
static void Forward_Tospace(int number_of_words)
{
  int i, mw;
  UWW mi;
  WamWord *p, *p2, *to;
  WamWord w;
  i = number_of_words;
  to = tospace + i;
  while(i-- > 0) {
    to--;
    mw = IndexToMarkWord(i);
    mi = IndexToMarkIndex(i);
    if(IsMoving(mw,mi)) {
      w = *to;
      p = UnTag_Stack(w);
      p2 = heap_start + *p;
      *to = Tag_Value(Tag_Of(w), p2);
    }
  }
}

/* Go through the marked bits and copy them from the heap to tospace. At the
   same time, compactify the "moving bits". */
static void Scan_and_Copy()
{
  WamWord *mbp, *to, *p;
  UWW bits, mi, mj;
  int i,j,mw;

  j = 0;
  to = tospace;
  mbp = markbits;
  for(;;) {
    bits = *mbp++;
    if(bits != 0) {
      if(mbp >= tospace_end) break;
      i = mbp - markbits - 1;
      mi = 1;
      p = heap_start + (i << LOG_WS);
      do {
	if(bits & 1) {
	  /* This particular word should be copied */
	  *to = *p;
	  /* And linked */
	  *p = to - tospace;
	  to++;
	  /* and then we should compress the moving bits */
	  mw = IndexToMarkWord(j);
	  mj = IndexToMarkIndex(j);
	  j++;
	  if(IsMoving(i,mi)) {
	    MarkMoving(mw,mj);
	  }
	  else {
	    MarkNonMoving(mw,mj);
	  }
	}
	p++;
	mi <<= 1;
	bits >>= 1;
      }while(bits != 0);
    }
  }
}
 

static WamWord *Updated_Heap_Marker(WamWord *p)
{
  int i, mw;
  UWW mi;

  i = PointerToIndex(p)-1;
  if(i < 0 || not_above < 0) {
    if(p < heap_start) { /* Which means p was saved prior to GC_Init */
      return p;
    }
    else {  /* Everything below is gone */
      not_above = -1;
      return heap_start;
    }
  }
  if(i >= not_above) { /* Nothing live between markers */
    return heap_start + heap_start[not_above] + 1;
  }
  /* Something live between markers -- check first word */
  mw = IndexToMarkWord(i);
  mi = IndexToMarkIndex(i);
  do {
    if(IsMarked(mw,mi)) {
      not_above = i;
      return heap_start + heap_start[i] + 1;
    }
    i--;
    mi >>= 1;
  }while(mi != 0);
  /* Nope, no hit. Run through earlier words. */
  while(--mw >= 0) {
    if(markbits[mw] != 0) { 
      mi = 1 << (WORD_SIZE - 1);
      i = (mw << LOG_WS) + WORD_SIZE - 1;
      for(;;) { /* OK, this *must* yield a hit */
	if(IsMarked(mw, mi)) {
	  not_above = i;
	  return heap_start + heap_start[i] + 1;
	}
	i--;
	mi >>= 1;
      }
    }
  }
  /* Nothing live below this marker */
  not_above = -1;
  return heap_start;
}


static void Mark_Trail(void)
{
  WamWord *tr, *trail, *adr;
  WamWord w;
  int nb;

  tr = TR;
  trail = Trail_Stack;
  while(tr > trail) {
    w = *--tr;
    adr = (WamWord *) (Trail_Value_Of(w));
    switch(Trail_Tag_Of(w)) 
      {
      case TUV:
	/* Actually, I believe that this should be accessible anyways, but
	   if it is it costs very little to mark it twice, and if I am wrong
	   and it is not accessible, then one need to compress the trail
	   in Forward_Trail, which is not done now (so it might cause strange
	   bugs on backtracking if there has been _two_ garbage collects...) */
	Mark_Root(Tag_REF(adr)); 
	break;
      case TOV:
	/* No clue (it is FD stuff), but I hope it is on CSTR stack ...*/
	tr--;
	break;
      case TMV:
	/* No clue (it is FD stuff), but I hope it is on CSTR stack ...*/
	nb = *--tr;
	tr -= nb;
	break;
      default:
	/* This seems dangerous to a GC, too... */
	break;
      }
  }
}

static void Forward_Trail(void)
{
  WamWord *tr, *trail, *adr;
  WamWord w;
  int nb;

  tr = TR;
  trail = Trail_Stack;
  while(tr > trail) {
    w = *--tr;
    adr = (WamWord *) (Trail_Value_Of(w));
    switch(Trail_Tag_Of(w)) 
      {
      case TUV:
	w = Forward_Root(Tag_REF(adr)); 
	adr = UnTag_Stack(w);
	*tr = Trail_Tag_Value(TUV, adr);
	break;
      case TOV:
	/* No clue (it is FD stuff), but I hope it is on CSTR stack ...*/
	tr--;
	break;
      case TMV:
	/* No clue (it is FD stuff), but I hope it is on CSTR stack ...*/
	nb = *--tr;
	tr -= nb;
	break;
      default:
	/* This seems dangerous to a GC, too... */
	break;
      }
  }  
}

static void Mark_Environment(WamWord *e, int n)
{
  int i;
  for(i = 0; i < n; i++) Mark_Root(Y(e,i));
}

static void Forward_Environment(WamWord *e, int n)
{
  int i;
  for(i = 0; i < n; i++) Y(e,i) = Forward_Root(Y(e,i));
}

static void Mark_Choicepoint(WamWord *b, int n)
{
  int i;
  for(i = 0; i < n; i++) Mark_Root(AB(b,i));
}

static void Forward_Choicepoint(WamWord *b, int n)
{
  int i;
  for(i = 0; i < n; i++) AB(b,i) = Forward_Root(AB(b,i));
  HB(b) = Updated_Heap_Marker(HB(b));
}

static void Mark_Locals(void)
{
  WamWord *curr, *prev, *e, *b, *eb, *stack;
  e = E;
  b = B;
  stack = Local_Stack;
  while(b > stack || e > stack) { /* There is something on the stack */
    if(b > e) { /* Top thing is a choicepoint */
      curr = b;
      b = BB(curr);
      eb = EB(curr);
      if(eb > e) e = eb; /* Some deallocated env. to bring back to life */
      prev = b >= e ? b : e;
      Mark_Choicepoint(curr, curr-prev-CHOICE_STATIC_SIZE);
    }
    else { /* Top thing is an environment */
      curr = e;
      e = EE(e);
      prev = b >= e ? b : e;
      Mark_Environment(curr, curr-prev-ENVIR_STATIC_SIZE);
    }
  }
}

static void Forward_Locals(void)
{
  WamWord *curr, *prev, *e, *b, *eb, *stack;
  e = E;
  b = B;
  stack = Local_Stack;
  while(b > stack || e > stack) { /* There is something on the stack */
    if(b > e) { /* Top thing is a choicepoint */
      curr = b;
      b = BB(curr);
      eb = EB(curr);
      if(eb > e) e = eb; /* Some deallocated env. to bring back to life */
      prev = b >= e ? b : e;
      Forward_Choicepoint(curr, curr-prev-CHOICE_STATIC_SIZE);
    }
    else { /* Top thing is an environment */
      curr = e;
      e = EE(e);
      prev = b >= e ? b : e;
      Forward_Environment(curr, curr-prev-ENVIR_STATIC_SIZE);
    }
  }  
}

static void Mark_Registers(int n)
{
  int i;

  for(i = 0; i < n; i++) Mark_Root(X(i));
}

static void Forward_Registers(int n)
{
  int i;
  for(i = 0; i < n; i++) X(i) = Forward_Root(X(i));
  /* H = Updated_Heap_Marker(H); /* The slower way of doing this */
  H = heap_start + count;        /* The faster way of doing this */
}

/* Of course, it doesn't swap them at all... */
static void Swap_Spaces(void)
{
  memcpy(heap_start, tospace, count * sizeof(WamWord));
}

static void Init_Vars(void)
{
  count = 0;
  memset(markbits, 0, (tospace_end - markbits - 1) * sizeof(WamWord));
  *tospace_end = 1; /* Sentinel for scanning -- speeds it (very) slightly */
  lis_sp = tospace;
  stc_sp = movingbits;
  not_above = PointerToIndex(H);
}

/* This is called to set up the buffers needed by garbage collection. */

Bool GC_Init(int max_fill_percent, int marg)
{
  unsigned long all, size, size_bb;

  if(marg == 0) marg = DEFAULT_MARGIN;
  else if(marg < 0) {
    marg = (Global_Size - 1) / DEFAULT_MARGIN_FRACTION;
  }
  margin = marg;
  memory_catch_B = B;
  num_catches = 0;
  heap_start = H;
  tospace_end = Global_Stack + Global_Size - 1;
  heap_trigger_out = tospace_end - margin;
  if(max_fill_percent >= 100 || max_fill_percent < -1) return FALSE;
  if(max_fill_percent == -1) max_fill_percent = DEFAULT_FILL;
  if(max_fill_percent == 0) {
    max_count = 0;
    heap_trigger_gc = tospace_end - margin;
    return TRUE;
  }
  all = tospace_end - heap_start;
  size_bb = ((all*4)/((max_fill_percent + 100)*WORD_SIZE + 200))*25;
  size = size_bb * WORD_SIZE;
  heap_trigger_gc = heap_start + size - margin;
  tospace = heap_end = heap_start + size;
  markbits = tospace_end - size_bb;
  movingbits = markbits - size_bb;
  max_count = all - size - 2*size_bb;  
  return TRUE;
}

Bool GC_Init_2(WamWord w, WamWord w2)
{
  int p, m;
  WamWord target;
  int tag;
  WamWord *dummy;
  Deref(w, target, tag, dummy);
  if(tag != INT) return FALSE;
  p = UnTag_INT(target);
  Deref(w2, target, tag, dummy);
  if(tag != INT) return FALSE;
  m = UnTag_INT(target);
  return GC_Init(p,m);
}

static Bool gc(int n)
{
  Init_Vars();
  Mark_Registers(n);
  if(count > max_count) return FALSE;
  Mark_Trail();
  if(count > max_count) return FALSE;
  Mark_Locals();
  if(count > max_count) return FALSE;
  Scan_and_Copy();
  Forward_Tospace(count);
  Forward_Registers(n);
  Forward_Trail();
  Forward_Locals();
  Swap_Spaces();
  return TRUE;
}

Bool GC_Set_Catcher(void)
{
  if(num_catches >= MAX_NUM_CATCHES) return FALSE;
  old_memory_catches[num_catches++] = memory_catch_B;
  memory_catch_B = B;
  return TRUE;
}

Bool GC_Reset_Catcher(void)
{
  if(num_catches <= 0) return FALSE;
  memory_catch_B = old_memory_catches[--num_catches];
  return TRUE;
}

Bool GC_Is_Out_Of_Memory(void)
{
  if(max_count < 0) return TRUE;
  else return FALSE;
}

#define MEM_OUT (H >= heap_trigger_out)
#define CAN_GC (H < heap_end)
#define WANT_GC (H >= heap_trigger_gc)

#define SHOULD_GC (WANT_GC && (CAN_GC || MEM_OUT))
    
Bool GC_GC(int arity)
{
  if(MEM_OUT) { /* Ouch! */
    max_count = -1;
    heap_trigger_gc = heap_trigger_out;
    B = memory_catch_B;
    return FALSE;
  }
  else if(CAN_GC && max_count > 0) { /* This is normal operation */
    int res;
    /* printf("trying to collect from %d used cells\n", PointerToIndex(H)); */
    res = gc(arity);
    if(res == FALSE) { /* not enough space reclaimed, turn GC off */
      heap_trigger_gc = H - margin;
      heap_end = H;
    } 
    else { /* We succeeded. In case we have just turned GC on, update */
      /* printf("down to %d live cells\n", PointerToIndex(H)); */
      heap_end = tospace;
      heap_trigger_gc = heap_end - margin;
    }
    return TRUE;
  }
  else if(H < tospace && max_count > 0) { 
    /* i.e. we have overrun heap_end, and yet we are 
       called -- presumably by user. Do some extra work! */
    heap_trigger_gc = H - 1;
    heap_end = tospace;
    return GC_GC(arity);
  }
  else { /* called by user (?) but we cannot run */
    return TRUE;
  }
}

Bool GC_GC_0(void) 
{
  return GC_GC(0);
}

Bool GC(int arity)
{
  if(SHOULD_GC) return GC_GC(arity);
  return TRUE;
}

Bool GC_0(void)
{
  if(SHOULD_GC) return GC_GC(0);
  else return TRUE;
}

Bool Pre_Jump(int arity)
{
  if(SHOULD_GC) return GC_GC(arity);
  return TRUE;
}

Bool Post_Allocate0(int arity)
{
  if(SHOULD_GC) return GC_GC(arity);
  return TRUE;
}

Bool Post_Allocate1(int arity)
{
  if(SHOULD_GC) return GC_GC(arity);  
  Y(E,0) = NOT_A_WAM_WORD;
  return TRUE;
}

Bool Post_Allocate2(int arity)
{
  if(SHOULD_GC) return GC_GC(arity);
  Y(E,0) = NOT_A_WAM_WORD;
  Y(E,1) = NOT_A_WAM_WORD;
  return TRUE;
}

Bool Post_Allocate3(int arity)
{
  if(SHOULD_GC) return GC_GC(arity);
  Y(E,0) = NOT_A_WAM_WORD;
  Y(E,1) = NOT_A_WAM_WORD;
  Y(E,2) = NOT_A_WAM_WORD;
  return TRUE;
}

Bool Post_AllocateN(int arity, int n)
{
  int i;

  if(SHOULD_GC) return GC_GC(arity);
  for(i = 0; i < n; i++) Y(E,i) = NOT_A_WAM_WORD;
  return TRUE;
}

void Post_Allocate_Lite(int n)
{
  int i;

  for(i = 0; i < n; i++) Y(E,i) = NOT_A_WAM_WORD;
}

